/////////////////////////////////////////////////////////////////////////////
//File     :  TankReader.java
//Author(s):  Chris (Episthene) Baillargeon
//Description: Read and Extract Tank Files. can (should) be run as a thread.

//copyright (C) 2005 Chris (Episthene) Baillargeon

//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//http://www.gnu.org/licenses/gpl.html

//----------------------------------------------------------------------------
//$Revision: 1.1               $Date: 2005/10/09  
//----------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////

package epi_extractor;
import java.io.*;
import java.nio.*;
import java.util.zip.*;
import java.nio.channels.*;
import java.util.concurrent.*;
import epi_tools.*;

public class TankReader implements Callable<String> 
{
	private TankHeader tankHeader;
	private File compressedTank;
	private String rootExtractFolder;
	private ByteBuffer bufferIndex;
	private int folderDirectoryOffset;
	private FileInputStream in;
	private FileChannel inChannel;
	
	//true to actually extract the files
	private boolean needToExtract = true;
	//true to decompress the files without extracting (if extracting = false), not really useful for now
	private boolean needToDecompress = true;
	//if false, Liquid files will be skipped
	private boolean needToExtractLiquid = true;
	
	//Reference to the Caller Object for the Error Callback
	private ITankCaller caller;
	
	//Variable for optional GUI Progress Bar
	private boolean isStarted = false;
	private String currentExtractedFile; 
	private int taskSize = 0;
	private int taskCurrent = 0;
	
	//Constructor
	public TankReader(String filIn, String dirOut, ITankCaller callerReference, boolean extractFiles, boolean decompress, boolean extractLiquid){

			compressedTank = new File(filIn);
			rootExtractFolder = dirOut;
			needToExtract = extractFiles;
			needToDecompress = decompress;
			needToExtractLiquid = extractLiquid;
			caller = callerReference;
	}
	
//Read the Tank and Extract	
	private void readTank() throws Exception {
		
		//Error Check for Valid Folder
    	File path = new File(rootExtractFolder);
    	if (!path.isDirectory() || !path.exists()){
    		throw new InvalidDirectoryException("Tank Extractor: The Directory chosen is not valid");
    	}   
    	
		//Error Check for Valid File
    	if (compressedTank.isDirectory() || !compressedTank.exists()){
    		throw new InvalidFileException("Tank Extractor: The File chosen is not valid");
    	}   
		
		in = new FileInputStream (compressedTank);
		inChannel = in.getChannel();
		
		//Get Header Info
		tankHeader  = new TankHeader(inChannel);

		//Error Check for Valid Tank
    	if (!tankHeader.getTankId().equalsIgnoreCase("TANK")){
    		throw new InvalidTankException("Tank Extractor: The file chosen is not a valid Tank file");
    	}
    	
    	isStarted = true;
		taskSize = tankHeader.getDirectoryOffset() - tankHeader.getDataOffset();
		
		//Get the Offset of the Folders. Skip the FirstDirectory with all Files and Folders
		
		//Get Index in memory
		bufferIndex = ByteBuffer.allocate(tankHeader.getIndexSize());
		bufferIndex.order(ByteOrder.LITTLE_ENDIAN);
		inChannel.read(bufferIndex, tankHeader.getDirectoryOffset());
		
//		Get Number of Files or Directories
		int numberOfDirectories = bufferIndex.getInt(0);
		
		//Get Offset to the Folders Directory
		folderDirectoryOffset = (numberOfDirectories * 4) + 4;
		
		//Call the recursive function that looks in all directories and extract
		readDirectory(folderDirectoryOffset, rootExtractFolder);
		
		//Close the Tank
		inChannel.close();
		in.close();
		
		currentExtractedFile = compressedTank.getName();
	}	

	//Recusively reads Directories and Create Them. Also calls the function to extract Files
	private void readDirectory(int filOrDir, String dirOut) throws Exception
	{
		int newOffset = filOrDir;
		int numberOfFiles;
		int nameLenght;
		String directoryName;
		int i = 0;
		int fileOrDirOffset;
		String dirPath = dirOut;
		
		//4 - Parent Offset
		//4 - Number of Files
		//8- TimeStamp?
		//2- Name Lenght
		//Name Lenght - Name
		//(0- 3) - Padding
		//(4* Number of Files) - Files Offsets
		
		numberOfFiles =  bufferIndex.getInt(newOffset + 4); 
		nameLenght =  bufferIndex.getShort(newOffset + 16);
		
		directoryName = Epi_Tank_Tools.getStringFromBuffer(bufferIndex, newOffset + 18, nameLenght);
		
		newOffset = newOffset + 18 + nameLenght;//(numberOfFiles * 4)
		//Add Padding
		newOffset = newOffset + Epi_Tank_Tools.getFileNullPadding(nameLenght);

//		Build the Directory Path Name
		if (directoryName.compareToIgnoreCase("") == 0)
		{}
		else
		{
			dirPath = dirOut + "/" +  directoryName;
		}
	
		//Create this directory
		if (needToExtract)
		{
			createDirectory(dirPath);
		}
		
		while(i < numberOfFiles)
		{
			fileOrDirOffset = bufferIndex.getInt(newOffset + (i * 4));
			
			if (fileOrDirOffset < tankHeader.getFileOffset() - tankHeader.getDirectoryOffset())
			{
				//This is a directory				
				readDirectory(fileOrDirOffset, dirPath);
			}
			else
			{
				//This is a file
				//Get File Index Header
				FileIndexHeader fil = new FileIndexHeader(fileOrDirOffset, bufferIndex);
								
				//Check if we need to decompress
				if (needToExtract || needToDecompress)
				{			
					currentExtractedFile = dirPath + "/" + fil.fileName;
					currentExtractedFile = currentExtractedFile.substring(rootExtractFolder.length() + 1);
					//C:\Java Projects\Extract/art/meshes/characters/evil_d_monsters/016/m_c_edm_016_pos_1.asp
					readFile(dirPath, fil);
					
				}
			}
			i++;
		}
	}
	
	//Read the file Info and uncompress
	private void readFile(String dirPath, FileIndexHeader fileIndex) throws Exception
	{
		int newOffset;
		int compressedLenght;
		int chunkOffset = 0;
		int chunkDecompressedLenght = 0;
		int chunkCompressedLenght = 0;
		int moreChunksMarker;
		int chunks = 0;
		int taskLenght;
		String filePath;
		
		newOffset = fileIndex.startOfChunksIndex;

		filePath = dirPath + "/" + fileIndex.fileName;
		
		if (fileIndex.compressionFormat == 1)
		{
			//Compressed
			//InflaterInputStream
			//4 - Compressed Lenght
			//4 - Maximum Chunk Size
			
			compressedLenght = bufferIndex.getInt(newOffset);
			moreChunksMarker = 16;
			taskLenght = compressedLenght;
			
			//Check if we need to extract Liquid Files. Need to be here for the TaskLenght
			//Liquid Files are all compressed
			if (!needToExtractLiquid && new AllLiquidFilesFilter().accept(new File(fileIndex.fileName)))
			{
				taskCurrent = taskCurrent + taskLenght;
				return;
			}
			//Put Chunked File in Memory
			
//System.out.println("Decompressed File Lenght: " + fileIndex.decompressedLenght);			
//System.out.println("Max Chunk Size: " + maxChunkSize);
//System.out.println("Relative First Chunk OffSet: " + fileIndex.relativeFirstChunkOffset);
//System.out.println("Name: " + fileIndex.fileName);
//System.out.println("File Compressed Lenght: " + compressedLenght);
		
			while (moreChunksMarker != 0)
			{
				//4 - Chunk Decompressed Lenght
				//4 - Chunk Compressed Lenght
				//4 - More Chunks Marker (0 for none, 16 for more)
				//4 - Chunk Offset
				chunkDecompressedLenght = bufferIndex.getInt(newOffset + 8 + (16 * chunks));
				chunkCompressedLenght = bufferIndex.getInt(newOffset + 12 + (16 * chunks));
				moreChunksMarker = bufferIndex.getInt(newOffset + 16 + (16 * chunks));
				chunkOffset = bufferIndex.getInt(newOffset + 20 + (16 * chunks)) + fileIndex.relativeFirstChunkOffset + tankHeader.getDataOffset();

//System.out.println("New OffSet: " + newOffset);
//System.out.println("chunkDecompressedLenght: " + chunkDecompressedLenght);
//System.out.println("chunkCompressedLenght: " + chunkCompressedLenght);	
//System.out.println("moreChunksMarker: " + moreChunksMarker);
//System.out.println("Chunk OffSet: " + chunkOffset);
				
				 //Check if the chunk is compressed. Some chunks are not for some reason.
				if (chunkDecompressedLenght != chunkCompressedLenght)
				{
//					Uncompress
					 Inflater decompresser = new Inflater();
					 
					//Store the Compressed data 
					ByteBuffer buffer = ByteBuffer.allocate(chunkCompressedLenght);	
					inChannel.read(buffer, chunkOffset);
						 
					 // Uncompress the uncompressed data
					 byte[] result = new byte[chunkDecompressedLenght];
					 decompresser.setInput(buffer.array());				 
				 	 decompresser.inflate(result);
					 decompresser.end();
					 
					 if (moreChunksMarker != 0)
					 {
						//Add the non-compressed end of chunk data
						buffer = ByteBuffer.allocate(moreChunksMarker);	
						inChannel.read(buffer, chunkOffset + chunkCompressedLenght);
						buffer.rewind();
						buffer.get(result, result.length - moreChunksMarker, moreChunksMarker);
					 }
//System.out.println("Compressed Chunked File: " + fileIndex.fileName);	
					 //Write the Compressed Chunk
					 if (chunks == 0)
					 {
						 writeFile(filePath, result, false);
					 }
					 else
					 {
						 writeFile(filePath, result, true);
					 } 
					result = null;	 
				}
				else
				{
//System.out.println("Uncompressed Chunked File: " + fileIndex.fileName);
					//Read and Write the non-compressed chunk
					ByteBuffer buffer = ByteBuffer.allocate(chunkDecompressedLenght);
					inChannel.read(buffer, chunkOffset);

					 if (chunks == 0)
					 {
						 writeFile(filePath, buffer.array(), false);
					 }
					 else
					 {
						 writeFile(filePath, buffer.array(), true);
					 }
				}
				chunks = chunks + 1;
			}
		}
		else
		{
//System.out.println("Copying: " + dirPath + "/" + fileIndex.fileName + " ->" );
//System.out.println("Decompressed File Lenght: " + fileIndex.decompressedLenght);			
//System.out.println("Relative First Chunk OffSet: " + fileIndex.relativeFirstChunkOffset);
//System.out.println("Name: " + fileIndex.fileName);
//System.out.println("Data Offset: " + tankHeader.getDataOffset());
//System.out.println("Offset: " + (fileIndex.relativeFirstChunkOffset + tankHeader.getDataOffset()));
//System.out.println("Uncompressed File: " + fileIndex.fileName);	

			taskLenght = fileIndex.decompressedLenght;
			//Not Chunked or Compressed
			writeFileDirect(filePath, fileIndex.relativeFirstChunkOffset + tankHeader.getDataOffset(), fileIndex.decompressedLenght);

		}
		taskCurrent = taskCurrent + taskLenght;		
	}
	
	//For uncompressed files
	private void writeFileDirect(String dirPath, int offset, int lenght) throws Exception
	{
		if (!needToExtract){return;}

		File fil = new File(dirPath);
		fil.createNewFile();
		FileChannel out = new FileOutputStream(dirPath).getChannel();

		//inChannel.position(offset);
		inChannel.transferTo(offset, lenght, out);

		out.close();
	}

	///For compressed files
	private void writeFile(String dirPath, byte[] in, boolean append) throws Exception
	{
		if (!needToExtract){return;}
	
		FileChannel out;
		
		if (append)
		{
			out = new FileOutputStream(dirPath, true).getChannel();
		}
		else
		{
			File fil = new File(dirPath);
			fil.createNewFile();
			out = new FileOutputStream(dirPath).getChannel();
		}
		
		out.write(ByteBuffer.wrap(in));	
		out.close();
		
	}
	
	private void createDirectory(String dirPath)
	{
		File newDir;
		
		newDir = new File(dirPath);
		newDir.mkdir();
	}
	
//	Thread method
	 public String call () throws Exception{

			try{
				//Only one Main method for now
				readTank();
			}catch(Exception e){
				caller.callback(e);
	            
			};		 
	    return "Extraction Complete: " + compressedTank.getAbsolutePath();
	  }
	
	
	//Methods to be invoked by a GUI Progress Bar
	public int getLengthOfTask()
	{
		return taskSize;
	}
	public int getCurrent()
	{
		return taskCurrent;
	}
	public String getMessage()
	{
		return currentExtractedFile;
	}
	
	//Tells that the Tank Reader is working
	public boolean isStarted()
	{
		return isStarted;
	}
	
	public void resetIsStarted()
	{
		isStarted  = false;
	}

}
	
